/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Point2D.h - definicao das classes Point2D e GraphicPoint2D.
 *
 *             link flag -lm
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    19.11.91  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _POINT2D_DEFINED
#define _POINT2D_DEFINED

#include <iostream.h>
#include <stdio.h>
#include "global.h"
#include "utilities.h"
#include "handle.h"
#include "World.h"
#include "Coords2D.h"
#include "Obj.h"
#include "basic.h"
#include "List.h"

/*
 * Observacao:
 *
 ******************************************************************************
 *
 *	bool left_turn(const Point2D &p, const Point2D &q) const;
 *	bool right_turn(const Point2D &p, const Point2D &q) const;
 *	bool colinear(const Point2D &p, const Point2D &q) const;
 *
 *	SideType which_side(const Point2D &p, const Point2D &q) const;
 *
 ******************************************************************************
 *
 *	As rotinas listadas acima  usam tecnicas especificas relacionadas 
 *      com coordenadas homogeneas para obter um alto grau de robustez. 
 *      As principais referencias que devem ser consultadas para conseguir 
 *	familiaridade com o assunto sao:
 *
 *	[1] Ahuja, D. V. and Coons S. A.
 *	    "Geometry for Construction and Display" IBM Systems Journal, 
 *          N. 3 & 4 1968
 *		
 *	[2] Newman W. M. and Sproull R. F
 *	    "Principles of Interactive Computer Graphics", McGraw-Hill
 *	    computer science series, 1973
 *
 *	[3] Harrington, S.
 *	    "Computer Graphics - A Programming Approach", 2nd edition
 *	    McGraw-Hill, 1987
 *
 *	[4] Rogers, D. F. and Adams J. A.
 *	    "Mathematical Elements for Computer Graphics", McGraw-Hill
 *          computer science series, 1976
 *
 ******************************************************************************
 *
 *   Nota: 
 *            Os metodos left_turn, right_turn ... funcionam da seguinte
 *            forma:
 *
 *            dados p e q, calcula o vetor v que define a reta direcionada
 *            que passa por eles da forma:
 *
 *            v = p x q (cross product)			pag. 472 [2]
 *
 *	      em seguida, para determinar de que lado de v um dado ponto PT
 *            esta' calcula-se d como sendo:
 *
 *            d = PT * v (dot product)			pag. 471 [2]
 *
 *            a posicao de PT com relacao a v e' dada pelo valor de d
 *
 *            = 0	PT esta' sobre v (colinear a p e q)     
 *            > 0	PT esta' `a esquerda da linha direcionada pq
 *	      < 0	PT esta' `a direita da linha direcionada pq
 *
 ******************************************************************************
 */

// forward declaration
class GraphicPoint2D;
 
class Point2D : public Obj {
  public:
    // construtores
    Point2D(char *s = "Point2D", ObjType t = Generic_t, bool v = false) : Obj(s, t, v) { setXY(0, 0); }
    Point2D(double x, double y) : Obj("Point2D") { setXY(x, y); }
    Point2D( Point2D &p) : Obj(p.what_is(), p.stype, p.volati) {	  
	if (p.at_infinity())
	  set_at_infinity(p.X(), p.Y());
	else
	  setXY(p.X(), p.Y());
    }
    
    //destrutor
    ~Point2D() {}

    // construtor virtual (simulado) que estabelece o relacionamento
    //
    //                  Point2D --> GraphicPoint2D
    
    GraphicObj* new_GraphicObj();
    
    // funcao para auto-reproducao de um objeto - necessario para possibilitar 
    // que a interface manipule objetos externos ao geolab (alem disso torna
    // mais homogeneo o tratamento de objetos conhecidos e desconhecidos)

    Obj* reproduct()  {
    	return new Point2D(*this);
    }
        
    // query & set functions using homogeneous coordinates
    virtual Point2D& operator = ( Point2D &p) {
        if (p.at_infinity())
          set_at_infinity(p.X(), p.Y());
        else
          setXY(p.X(), p.Y());
        return *this;
    }
        
    double X()  { return hp.X(); }
    double Y()  { return hp.Y(); }

    void setXY(double x, double y) { hp.setXY(x, y); }

    void set_at_infinity(double x, double y) { hp.set_at_infinity(x, y); }
        
    bool at_infinity()  { return hp.at_infinity(); }

    // functions to compare points lexicographically, circulary given a
    // reference point, etc
    bool lessX( Point2D &q)   { return less(X(), q.X()); }
    bool lessY( Point2D &q)   { return less(Y(), q.Y()); }
    bool equalX( Point2D &q)  { return ::equal(X(), q.X()); }
    bool equalY( Point2D &q)  { return ::equal(Y(), q.Y()); }
    
    bool lexicoXless( Point2D &q)   { return lessX(q) || (equalX(q) && lessY(q)); }
    bool lexicoYless( Point2D &q)   { return lessY(q) || (equalY(q) && lessX(q)); }
    bool equal( Point2D &q)         { return equalX(q) && equalY(q); }
    bool operator == ( Point2D &q)  { return equalX(q) && equalY(q); }
    
    bool circular_less( Point2D &q,  Point2D &ref_point) ;
    bool circular_equal( Point2D &q,  Point2D &ref_point) ;

    // geometric functions
    bool left_turn( Point2D &p,  Point2D &q)  {
    	return ::less(0, hp * cross(p.hp, q.hp));
    }
    
    bool right_turn( Point2D &p,  Point2D &q)  {
    	return ::less(hp * cross(p.hp, q.hp), 0);
    }
    
    bool colinear( Point2D &p,  Point2D &q)  {
    	return ::equal(hp * cross(p.hp, q.hp), 0);
    }

    SideType which_side( Point2D &p,  Point2D &q)  {
    	if (colinear(p, q))
    	  return Colinear;
    	else
      	  if (left_turn(p, q))
    	    return Left;
    	  else
    	    return Right;
    }

    /*
     *
     *   The pre-condition following geometric functions is that no point 
     *   is at infinity. This is not verified at all.
     *
     */

    double area( Point2D &p,  Point2D &q)  {
    	return (p.X() * q.Y()) + (p.Y() * X()) + (q.X() * Y()) - (q.Y() * X()) - (p.Y() * q.X()) - (p.X() * Y());
    }
         
    inCircleType inCircle( Point2D &a,  Point2D &b,  Point2D &c) ;
    
    double e_distance( Point2D &q) ;
    double e_distance_sqr( Point2D &q) ;

    double angle( Point2D &q) ;    

    // inherited methods redefinition
    void load(FILE *f) { Obj::load(f); hp.load(f); }
    void save(FILE *f) { Obj::save(f); hp.save(f); }

    ObjType type()  { return Point2D_t; }
    
    void translate(double to_x, double to_y) {
        if (at_infinity())
          set_at_infinity(X() + to_x, Y() + to_y);
        else
          setXY(X() + to_x, Y() + to_y);
    }

    void rotate(double orig_x, double orig_y, double angle) {
	double cs = cos(angle), sn = sin(angle);	
	setXY(orig_x+(X()-orig_x)*cs-(Y()-orig_y)*sn,orig_y+(X()-orig_x)*sn+(Y()-orig_y)*cs);
    }

    
    void print()  { cout << "Point2D at (" << X() << "," << Y() << ")" << endl; }

    void unfill_handles(World *w, bool show_handle = true) ;
    void unfill_circular_handles(World *w, bool show_handle = true) ;
    void draw_handles(World *w, GC gc, bool highlight = false, bool show_handle = true) ;
    void draw_circular_handles(World *w, GC gc, bool highlight = false, bool show_handle = true) ;
    void draw(World *w, GC gc) ;

    bool select_test(double x, double y)  {
    	if (at_infinity())
    	  return false;
    	else
          return select_handle_test(hp.X(), hp.Y(), x, y);
    }
    
    bool select_test(double x1, double y1, double x2, double y2)  {
    	if (at_infinity())
    	  return false;
    	else
       	  return select_handle_test(hp.X(), hp.Y(), x1, y1, x2, y2);
    }

    double distance_to_point(double x, double y)  {
    	return e_distance(Point2D(x, y));
    }


    // Methods for information retrieval.
    virtual bool get_points_implemented() { return true; }
    virtual list(Obj_ptr) get_points()  {
      list(Obj_ptr) Objp;
      Point2D* p_ptr = new Point2D(*this); assert(p_ptr);
      Objp.append( p_ptr );
      return Objp;
    }

  protected:
    PlaneCoords hp;
};

/*
 *
 * GraphicPoint2D
 *
 *
 */

class GraphicPoint2D : public GraphicObj {
  public:
    // rutores
    GraphicPoint2D() : GraphicObj(new Point2D()) {}
    GraphicPoint2D(Point2D *p) : GraphicObj(p) {}
    
    //destrutor
    ~GraphicPoint2D() {}
    
    // inherited methods redefinition
    void int_start(World *, double initial_X, double initial_Y, Int_Mode);
    void int_change(World *, double new_X, double new_Y, Int_Mode);
    void int_end(World *, double final_X, double final_Y, Int_Mode);
    
    void resize_start(World *, double X, double Y, Int_Mode);
    void resize_change(World *, double X, double Y, Int_Mode);
    void resize_end(World *, double X, double Y, Int_Mode);

    // this is because they are points, ah ? so what ? try using...
    virtual void unfill_handles(World *w) {
    	((Point2D *)obj)->unfill_circular_handles(w, show_handle);
    }
    
    virtual void draw_handles(World *w) {
    	((Point2D *)obj)->draw_circular_handles(w, gch, is_selected, show_handle);
    }
    
    virtual void draw(World *w, DrawMode dm);
 
    virtual bool where_draw_string (double &x, double &y);
};

typedef Point2D* Point2D_ptr;
typedef GraphicPoint2D* GraphicPoint2D_ptr;

#ifndef _Point2D_ptr_list
#define _Point2D_ptr_list
declare (list,Point2D_ptr)
#endif ~_Point2D_ptr_list;

// Point related functions (that don't fit naturally in Point2D class)
Point2D gravitycentre( Point2D &a,  Point2D &b,  Point2D &c) ;
Point2D incentre( Point2D &a,  Point2D &b,  Point2D &c, double &radius) ;
Point2D circumcentre( Point2D &a,  Point2D &b,  Point2D &c) ;
Point2D baricentre( Point2D &a,  Point2D &b,  Point2D& c) ;
Point2D middle( Point2D &p,  Point2D &q) ;

Point2D get_max_X( list(Point2D_ptr) &Points) ;
Point2D get_max_Y( list(Point2D_ptr) &Points) ;
Point2D get_min_X( list(Point2D_ptr) &Points) ;
Point2D get_min_Y( list(Point2D_ptr) &Points) ;

list_item it_get_max_X( list(Point2D_ptr) &Points) ;
list_item it_get_max_Y( list(Point2D_ptr) &Points) ;
list_item it_get_min_X( list(Point2D_ptr) &Points) ;
list_item it_get_min_Y( list(Point2D_ptr) &Points) ;

Point2D lower_angle_cw( list(Point2D_ptr) &Points,  Point2D &ref) ;
Point2D lower_angle_ccw( list(Point2D_ptr) &Points,  Point2D &ref) ;

list_item it_lower_angle_cw( list(Point2D_ptr) &Points,  Point2D &ref) ;
list_item it_lower_angle_ccw( list(Point2D_ptr) &Points,  Point2D &ref) ;

void remove_duplicated_points(list(Point2D_ptr) &Points);
void remove_colinear_points(list(Point2D_ptr) &Points);

// functions to be passed to a generic sort function:

CompType lexico_Xcomp( Point2D &p1,  Point2D &p2) ;
CompType lexico_Ycomp( Point2D &p1,  Point2D &p2) ;
CompType lexico_Xcomp_ptr( Point2D_ptr &p1,  Point2D_ptr &p2) ;
CompType lexico_Ycomp_ptr( Point2D_ptr &p1,  Point2D_ptr &p2) ;

void set_circular_comp_reference( Point2D &p) ;

CompType circular_comp( Point2D &p1,  Point2D &p2) ;
CompType circular_comp_ref( Point2D &p1,  Point2D &p2,  Point2D &ref) ;
CompType circular_comp_ptr( Point2D_ptr &p1,  Point2D_ptr &p2) ;
CompType circular_comp_ptr_ref( Point2D_ptr &p1,  Point2D_ptr &p2,  Point2D_ptr &ref) ;

#endif ~_POINT2D_DEFINED;


