/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * ConvexHull.C - definicao da superclasse de algoritmos para a construcao de
 *                envoltorias convexas para conjuntos de pontos. Implementacao
 *                dos algoritmos Jarvis' March e Graham's Scan.
 * 
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    24.04.92  09:30 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "ConvexHull.h"

// includes para animacao

#include "Misc.h"
#include "Circle.h"
#include "Segment2D.h"
#include "Ray2D.h"

// variaveis para animacao

static GraphicCircle *anim_c;
static GraphicSegment2D *anim_s1;
static GraphicSegment2D *anim_s2;
static GraphicSegment2D *anim_s3;
static GraphicRay2D *anim_r;

#define anim_circle	((Circle *)anim_c->obj)
#define anim_segment1	((Segment2D *)anim_s1->obj)
#define anim_segment2	((Segment2D *)anim_s2->obj)
#define anim_segment3	((Segment2D *)anim_s3->obj)
#define anim_ray	((Ray2D *)anim_r->obj)

void convex_hull_animation_init(World *w) {
  static bool FIRST = true;
  
  if (FIRST) {
    anim_c = new GraphicCircle();
    anim_s1 = new GraphicSegment2D();
    anim_s2 = new GraphicSegment2D();
    anim_s3 = new GraphicSegment2D();
    anim_r = new GraphicRay2D();
    FIRST = false;
  }

  anim_c->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
  anim_s1->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
  anim_s2->set_all(w, FOREGROUND, BACKGROUND, 3, 0, false);
  anim_s3->set_all(w, FOREGROUND, BACKGROUND, 2, 1, false);
  anim_r->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
}

/*
 * To determinate if a given simple polygon is ordered counterclockwise, we
 * use the winding number algorithm.
 *
 */
 
static bool simple_polygon_ordered_counterclockwise( list(Point2D_ptr) &points) {
    if (points.size() > 2) {
      int wn, colinear = 0;
      list_item a, b, c;
      Point2D ref;
      SideType s, t;
      
      a = points.first(); b = points.succ(a); c = points.succ(b);
      ref = gravitycentre(*points[a], *points[b], *points[c]);
      wn = winding_number(points, ref);
      
      if (wn == 0) {
      	if (points[a]->left_turn(*points[b], *points[c]))
      	  s = Right;
      	else
      	  s = Left;
      	  
      	do {
       	  a = points.cyclic_succ(a); b = points.cyclic_succ(a); c = points.cyclic_succ(b);
       	  t = points[a]->which_side(*points[b], *points[c]);
       	  if (t == Colinear)
       	    colinear++;
        } while (t != s && colinear < points.size());
    
        if (colinear == points.size())
          return true;
          
        ref = gravitycentre(*points[a], *points[b], *points[c]);
        wn = winding_number(points, ref);
      }
      
      return (wn < 0);
    }
    else
      return true;
}
 
/*
 * Convex Hull of polygons
 *
 */
 
list(Point2D_ptr) ch_generic_polygon(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    return graham_scan(points, w, animation_depth, reference_depth);
}

/*
 * Convex Hull of a Simple Polygon in O(n) - implementation of Lee's algorithm
 * (reference: D. T. Lee - "On finding the convex hull of a simple polygon",
 *             International Journal of Computer and Information Sciences,
 *             v.12 pg.87-98, 1983)
 *
 */
 
void update(list(Point2D_ptr) &stack,  list(Point2D_ptr) &polygon, list_item v, Point2D &l1, Point2D &l2, World *w = 0, int animation_depth = 0, int reference_depth = 0) {
    list_item vl;

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      anim_ray->set(l1, l2);
      anim_r->draw(w, XorMode);
    ANIMATION_END(w, animation_depth, reference_depth)

    do {            
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_segment2->set(*stack.head(), *stack[stack.succ(stack.first())]);
        anim_s2->draw(w, XorMode);
      ANIMATION_END(w, animation_depth, reference_depth)
      
      stack.pop();		// remove the top vertex
      vl = stack.first();
      
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        if (vl != stack.last()) {
          anim_segment3->set(*stack[stack.succ(vl)], *stack[vl]);
          anim_s3->draw(w, XorMode);
          frame_stop_mark(w);
          anim_s3->draw(w, XorMode);
        }
      ANIMATION_END(w, animation_depth, reference_depth)
      
    } while (vl != stack.last() && !polygon[v]->left_turn(*stack[stack.succ(vl)], *stack[vl]));
   
    l1 = *stack.head();
    l2 = *polygon[v];

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      anim_ray->set(l1, l2);
      anim_r->draw(w, XorMode);
      frame_stop_mark(w);
    ANIMATION_END(w, animation_depth, reference_depth)
}

void close_lobe(list(Point2D_ptr) &stack,  list(Point2D_ptr) &polygon, list_item &v, Point2D &l1, Point2D &l2, World *w = 0, int animation_depth = 0, int reference_depth = 0) {
    list_item vl = polygon.cyclic_succ(v);

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      anim_circle->set(polygon[v]->X(), polygon[v]->Y(), 10);
      anim_c->draw(w, XorMode);
      anim_circle->set(polygon[vl]->X(), polygon[vl]->Y(), 10);
      anim_c->draw(w, XorMode);
      anim_segment3->set(*polygon[v], *polygon[polygon.pred(v)]);
      anim_s3->draw(w, XorMode);
      frame_stop_mark(w);
      anim_s3->draw(w, XorMode);
    ANIMATION_END(w, animation_depth, reference_depth)
      
    if (!polygon[vl]->right_turn(l1, l2) && polygon[vl]->left_turn(*polygon[v], *polygon[polygon.pred(v)])) {
      // vl is in region R0, case 2a
      do {

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(polygon[vl]->X(), polygon[vl]->Y(), 10);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)

        vl = polygon.cyclic_succ(vl);

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(polygon[vl]->X(), polygon[vl]->Y(), 10);
          anim_c->draw(w, XorMode);
          frame_stop_mark(w);
        ANIMATION_END(w, animation_depth, reference_depth)

      } while (!polygon[vl]->right_turn(l1, l2));

      v = vl;

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_ray->set(l1, l2);
        anim_r->draw(w, XorMode);
      ANIMATION_END(w, animation_depth, reference_depth)

      if (stack.size() > 1) {
        l1 = *stack[stack.succ(stack.first())];
        l2 = *stack.head();
      }
      else {
      	l1 = *stack.head();
      	l2 = *polygon[v];

        stack.push(polygon[v]);

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_segment2->set(*stack.head(), *stack[stack.succ(stack.first())]);
          anim_s2->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)
      }

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_ray->set(l1, l2);
        anim_r->draw(w, XorMode);
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_depth)
    }
    else {
      stack.push(polygon[v]);

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_segment2->set(*stack.head(), *stack[stack.succ(stack.first())]);
        anim_s2->draw(w, XorMode);
      ANIMATION_END(w, animation_depth, reference_depth)

      v = vl;
    }
}

list(Point2D_ptr) ch_simple_polygon(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    list(Point2D_ptr) result;
    
    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      convex_hull_animation_init(w);  
    ANIMATION_END(w, animation_depth, reference_depth)

    remove_colinear_points(points);
    
    if (points.size() > 2) {
      // 1st - assert that the polygon is ordered counterclockwise
      if (!simple_polygon_ordered_counterclockwise(points)) {
        list(Point2D_ptr) aux;
        Point2D_ptr p;
      
        forall(p, points) {
          aux.push(p);
        }
        points.clear();
        forall(p, aux) {
          points.append(p);
        }
      }
    
      // 2nd - points.head() must be the point with smallest Y coordinate
      list_item low_y = it_get_min_Y(points);
      while (low_y != points.first()) {
      	Point2D_ptr aux = points.pop();
      	points.append(aux);
      }
    
      // now the algorithm
      list(Point2D_ptr) stack;
      list_item next;
      Point2D l1, l2;

      stack.push(points.head());			// v0 pushed
      stack.push(points[points.succ(points.first())]);	// v1 pushed

      next = points.cyclic_succ(points.succ(points.first()));
      
      l1 = *points.head();
      l2 = *points[points.succ(points.first())];

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_ray->set(l1, l2);
        anim_r->draw(w, XorMode);
        anim_segment2->set(*stack.head(), *stack[stack.succ(stack.first())]);
        anim_s2->draw(w, XorMode);
        anim_circle->set(points[next]->X(), points[next]->Y(), 10);
        anim_c->draw(w, XorMode);
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_depth)
            
      while (next != points.first()) {
      	if (!points[next]->left_turn(l1, l2)) {
      	  update(stack, points, next, l1, l2, w, animation_depth, reference_depth);
      	  close_lobe(stack, points, next, l1, l2, w, animation_depth, reference_depth);
      	}
      	else {

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment3->set(*stack.head(), *points.head());
            anim_s3->draw(w, XorMode);
            frame_stop_mark(w);
            anim_s3->draw(w, XorMode);
          ANIMATION_END(w, animation_depth, reference_depth)
      
      	  if (!points[next]->right_turn(*stack.head(), *points.head()))
            // next in R1
      	    do {
      	      
      	      ANIMATION_BEGIN(w, animation_depth, reference_depth)
                anim_circle->set(points[next]->X(), points[next]->Y(), 10);
                anim_c->draw(w, XorMode);
      	      ANIMATION_END(w, animation_depth, reference_depth)
      	      
      	      next = points.cyclic_succ(next);

      	      ANIMATION_BEGIN(w, animation_depth, reference_depth)
                anim_circle->set(points[next]->X(), points[next]->Y(), 10);
                anim_c->draw(w, XorMode);
                frame_stop_mark(w);
      	      ANIMATION_END(w, animation_depth, reference_depth)
      	      
              ANIMATION_BEGIN(w, animation_depth, reference_depth)
                anim_segment3->set(*stack.head(), *points.head());
                anim_s3->draw(w, XorMode);
                frame_stop_mark(w);
                anim_s3->draw(w, XorMode);
              ANIMATION_END(w, animation_depth, reference_depth)
      
      	    } while (next != points.first() && !points[next]->right_turn(*stack.head(), *points.head()));
          else {
            // next in R2
      	    Point2D ll1 = *stack.head(), ll2 = *points[next]; 

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_ray->set(l1, l2);
              anim_r->draw(w, XorMode);
              anim_ray->set(ll1, ll2);
              anim_r->draw(w, XorMode);
              frame_stop_mark(w);
            ANIMATION_END(w, animation_depth, reference_depth)
            
      	    close_lobe(stack, points, next, ll1, ll2, w, animation_depth, reference_depth);
      	    l1 = ll1; l2 = ll2;
      	  }
      	}
      }
      
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_ray->set(l1, l2);
        anim_r->draw(w, XorMode);
        anim_circle->set(points[next]->X(), points[next]->Y(), 10);
        anim_c->draw(w, XorMode);
        anim_segment2->set(*stack.tail(), *stack[stack.first()]);
        anim_s2->draw(w, XorMode);
      ANIMATION_END(w, animation_depth, reference_depth)
            
      Point2D_ptr p;
      forall(p, stack) {
      	result.push(p);
      }
    }
    else {
      Point2D_ptr p;
      forall(p, points) {
        result.push(p);
      }
    }
    
    return result;
}

list(Point2D_ptr) ch_starshaped_polygon(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    list(Point2D_ptr) remainder, result;
    list_item vi, vi1, vi2;
    SideType side, scomp;
        
    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      convex_hull_animation_init(w);  
    ANIMATION_END(w, animation_depth, reference_depth)

    if (simple_polygon_ordered_counterclockwise(points))
      scomp = Right;
    else
      scomp = Left;
      
    if (points.size() > 2) {
      vi = points.first();

      do {
        vi1 = points.cyclic_succ(vi);
        vi2 = points.cyclic_succ(vi1);

        ANIMATION_BEGIN(w, animation_depth, reference_depth)
          anim_circle->set(points[vi]->X(), points[vi]->Y(), 6);
          anim_c->draw(w, XorMode);
          anim_circle->set(points[vi1]->X(), points[vi1]->Y(), 10);
          anim_c->draw(w, XorMode);
          anim_circle->set(points[vi2]->X(), points[vi2]->Y(), 6);
          anim_c->draw(w, XorMode);
    
          frame_stop_mark(w);

          anim_circle->set(points[vi]->X(), points[vi]->Y(), 6);
          anim_c->draw(w, XorMode);
          anim_circle->set(points[vi1]->X(), points[vi1]->Y(), 10);
          anim_c->draw(w, XorMode);
          anim_circle->set(points[vi2]->X(), points[vi2]->Y(), 6);
          anim_c->draw(w, XorMode);
        ANIMATION_END(w, animation_depth, reference_depth)
    
        side = points[vi]->which_side(*points[vi1], *points[vi2]);
   	if (side == scomp || (side == Colinear && vi2 != vi)) {

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment1->set(*points[vi], *points[vi1]);
            anim_s1->draw(w, XorMode);
            anim_segment1->set(*points[vi1], *points[vi2]);
            anim_s1->draw(w, XorMode);
            anim_segment1->set(*points[vi], *points[vi2]);
            anim_s1->draw(w, XorMode);
          ANIMATION_END(w, animation_depth, reference_depth)

          remainder.append(points.del_item(vi1));

    	  // faz backtrack somente se nao estiver no ponto de partida
      	   
      	  if (vi != points.first())
      	    vi = points.pred(vi);
      	}
      	else
      	  vi = points.succ(vi);
      } while (vi1 != points.first());
    }

    result = points;
    points = remainder;

    return(result);
}

/*
 * Convex Hull for point sets
 *
 */
 
list(Point2D_ptr) jarvis_march(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    list(Point2D_ptr) hull;
    list_item low_ccw, low_y;
    
    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      convex_hull_animation_init(w);  
    ANIMATION_END(w, animation_depth, reference_depth)

    if (!points.empty()) {
      if (animation(animation_depth+1, reference_depth))
        low_y = it_get_min_Y(points, w, animation_depth+1, reference_depth);
      else
        low_y = it_get_min_Y(points);
        
      hull.append(points[low_y]);
      points.append(points[low_y]);
      points.del_item(low_y);
      
      while (!points.empty()) {
      	if (animation(animation_depth+1, reference_depth))
       	  low_ccw = it_lower_angle_ccw(points, *hull.tail(), w, animation_depth+1, reference_depth);
        else
       	  low_ccw = it_lower_angle_ccw(points, *hull.tail());
        
      	if (*points[low_ccw] == *hull.head())
      	  break;
      	else {

          ANIMATION_BEGIN(w, animation_depth, reference_depth)
            anim_segment1->set(*hull.tail(), *points[low_ccw]);
            anim_s1->draw(w, XorMode);
            frame_stop_mark(w);
          ANIMATION_END(w, animation_depth, reference_depth)
      	  
      	  hull.append(points[low_ccw]);
      	}
      	points.del_item(low_ccw);
      }
      
      points.del_item(points.last());

      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        anim_segment1->set(*hull.head(), *hull.tail());
        anim_s1->draw(w, XorMode);
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_depth)
      	  
    }

    remove_colinear_points(hull);
    return hull;
}

list(Point2D_ptr) graham_scan(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    list(Point2D_ptr) result;
        
    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      convex_hull_animation_init(w);  
    ANIMATION_END(w, animation_depth, reference_depth)

    if (!points.empty()) {
      // ordena pontos circularmente (lexicograficamente) tomando
      // como ponto de referencia o de menor coordenada Y

      Point2D *ref;  
      if (animation(animation_depth+1, reference_depth))
        ref = points[it_get_min_Y(points, w, animation_depth+1, reference_depth)];
      else
        ref = points[it_get_min_Y(points)];
        
      set_circular_comp_reference(*ref);
      points.sort((int(*)(Point2D_ptr&, Point2D_ptr&))circular_comp_ptr);
      
      ANIMATION_BEGIN(w, animation_depth, reference_depth)
        list_item it;
        forall_list_items(it, points) {
          anim_segment1->set(*points[it], *points[points.cyclic_succ(it)]);
          anim_s1->draw(w, XorMode);
        }
        frame_stop_mark(w);
      ANIMATION_END(w, animation_depth, reference_depth)

      result = ch_starshaped_polygon(points, w, animation_depth, reference_depth);
    }
    
    return(result);
}

list(Polygon_ptr) convex_layers_jarvis(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    list(Polygon_ptr) result;
    Polygon_ptr p;

    while (!points.empty()) {
      p = new Polygon(jarvis_march(points, w, animation_depth, reference_depth));
      p->subtype(Convex_t); p->closed(true);
      result.append(p);
    }
    
    return result;
}

list(Polygon_ptr) convex_layers_graham(list(Point2D_ptr) &points, World *w, int animation_depth, int reference_depth) {
    list(Polygon_ptr) result;
    Polygon_ptr p;

    while (!points.empty()) {
      p = new Polygon(graham_scan(points, w, animation_depth, reference_depth));
      p->subtype(Convex_t); p->closed(true);
      result.append(p);
    }
    
    return result;
}

// Modelagem

list(Point2D_ptr) extract_points(const SelectedObjects &SO, const GraphicObjects &GO) {
    list(Point2D_ptr) res;
    list_item it;
    forall(it, SO) {			 	// is_able_to_handle deve ser
      res.push((Point2D *)GO[it]->obj);		// chamado antes para certificar
    }						// que sao todos pontos
    return res;
}

bool ConvexHull::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    list_item it;
    
    forall(it, SO) {
      if (GO[it]->type() != Point2D_t) {
      	SO.init_iterator();
      	return false;
      }
    }
    
    return true;
}

ResultObjects JarvisMarch::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Point2D_ptr) hull;
    
    if (!SO.empty()) {
      hull = jarvis_march(extract_points(SO, GO), w, animation_depth, reference_depth);

      if (!hull.empty()) {
        Polygon *p = new Polygon(hull);
      	p->subtype(Convex_t); p->closed(true);
      	result.push(p);
      }
    }
      
    return result;
}

ResultObjects GrahamScan::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Point2D_ptr) hull;
    
    if (!SO.empty()) {
      hull = graham_scan(extract_points(SO, GO), w, animation_depth, reference_depth);

      if (!hull.empty()) {
        Polygon *p = new Polygon(hull);
      	p->subtype(Convex_t); p->closed(true);
      	result.push(p);
      }
    }
      
    return result;
}

ResultObjects ConvexLayersJarvis::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Polygon_ptr) layers;
    Polygon_ptr p;
    
    if (!SO.empty()) {
      layers = convex_layers_jarvis(extract_points(SO, GO), w, animation_depth, reference_depth);

      forall(p, layers) {
      	result.push(p);
      }
    }
      
    return result;
}

ResultObjects ConvexLayersGraham::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Polygon_ptr) layers;
    Polygon_ptr p;
    
    if (!SO.empty()) {
      layers = convex_layers_graham(extract_points(SO, GO), w, animation_depth, reference_depth);

      forall(p, layers) {
      	result.push(p);
      }
    }
      
    return result;
}

bool GenericPolygonHull::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    return (SO.size() == 1 && GO[SO.head()]->type() == Polygon_t);
}

ResultObjects GenericPolygonHull::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Point2D_ptr) points;
    
    if (!SO.empty()) {
      points = ((Polygon *)GO[SO.head()]->obj)->copy_vertices();    
      Polygon *p = new Polygon(ch_generic_polygon(points, w, animation_depth, reference_depth));
      p->subtype(Convex_t); p->closed(true);
      result.push(p);
    }
      
    return result;
}

bool SimplePolygonHull::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    return (SO.size() == 1 && 
            GO[SO.head()]->type() == Polygon_t && 
            (GO[SO.head()]->subtype() == Simple_t || GO[SO.head()]->subtype() == StarShaped_t || GO[SO.head()]->subtype() == Convex_t));
}

ResultObjects SimplePolygonHull::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Point2D_ptr) points;
        
    if (!SO.empty()) {
      points = ((Polygon *)GO[SO.head()]->obj)->copy_vertices();    
      Polygon *p = new Polygon(ch_simple_polygon(points, w, animation_depth, reference_depth));
      p->subtype(Convex_t); p->closed(true);
      result.push(p);
    }
      
    return result;
}

bool StarShapedPolygonHull::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    return (SO.size() == 1 && 
            GO[SO.head()]->type() == Polygon_t && 
            (GO[SO.head()]->subtype() == StarShaped_t || GO[SO.head()]->subtype() == Convex_t));
}

ResultObjects StarShapedPolygonHull::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Point2D_ptr) points;
        
    if (!SO.empty()) {
      points = ((Polygon *)GO[SO.head()]->obj)->copy_vertices();    
      Polygon *p = new Polygon(ch_starshaped_polygon(points, w, animation_depth, reference_depth));
      p->subtype(Convex_t); p->closed(true);
      result.push(p);
    }
      
    return result;
}

Algorithm* jarvis_march_link() {
    return new JarvisMarch();
}

Algorithm* graham_scan_link() {
    return new GrahamScan();
}

Algorithm* convex_layers_jarvis_link() {
    return new ConvexLayersJarvis();
}

Algorithm* convex_layers_graham_link() {
    return new ConvexLayersGraham();
}

Algorithm* generic_polygon_hull_link() {
    return new GenericPolygonHull();
}

Algorithm* simple_polygon_hull_link() {
    return new SimplePolygonHull();
}

Algorithm* starshaped_polygon_hull_link() {
    return new StarShapedPolygonHull();
}






