/*
 * This code has been modified for compatibility with other
 * source codes of GeoLab.
 */

/*******************************************************************************
+
+  LEDA  2.0.1                                                7/3/1991
+
+
+  graph.h
+
+
+  Copyright (c) 1991  by  Max-Planck-Institut fuer Informatik
+  Im Stadtwald, 6600 Saarbruecken, FRG     
+  All rights reserved.
+ 
*******************************************************************************/



#ifndef GRAPHH
#define GRAPHH

/*------------------------------------------------------------------------------

    Graph

    Stefan Naeher
    FB10 Informatik 
    Universitaet des Saarlandes 
    6600 Saarbruecken

    September 1988
    
------------------------------------------------------------------------------*/


#include "basic.h"
#include "List.h"


//------------------------------------------------------------------------------
/* declare node, edge, list(node) and list(edge)                              */
//------------------------------------------------------------------------------

class i_node;
class i_edge;

// node = pointer to internal node
typedef i_node* node;

// edge = pointer to internal edge
typedef i_edge* edge;


/* nodelist = doubly linked list of nodes */
declare(list,node)

/* list(edge) = doubly linked list of edges */
declare(list,edge)



//------------------------------------------------------------------------------
// class i_node: internal representation of nodes
//------------------------------------------------------------------------------

class i_node {  

   friend class graph;
   friend class ugraph;
   friend class planar_map;
   
   friend graph* graph_of(edge);

   graph* g;             // pointer to graph 
   int i_name;           // internal name (index)  
   list_item loc;        // location in V
   int indegree;        
   ent contents;         // user defined pointer-type
   list(edge) adj_edges;   /* list of adjacent edges */ 

   i_node(ent i)
   { 
     contents=i;
     i_name = -1; 
     g = 0; 
     loc = 0; 
     indegree = 0;
   }

   OPERATOR_NEW(10)  
   OPERATOR_DEL(10)

friend graph* graph_of(node v)          { return v->g;         }
friend int    indeg(node v)             { return v->indegree;  }
friend int    outdeg(node v)            { return v->adj_edges.length(); }
friend int    degree(node v)            { return v->adj_edges.length(); }
friend int    index(node v)             { return v->i_name;    }

/* space :  5 words + 1 list = 20 + 20 bytes = 40 bytes */
};


//------------------------------------------------------------------------------
// class i_edge: internal representation of edges
//------------------------------------------------------------------------------

class i_edge { 
   
   friend class graph;
   friend class ugraph;
   friend class planar_map;

   friend graph* graph_of(edge);

   int i_name;           // internal name (index)  
   node s;               // edge from "source" 
   node t;               // to "target"
   list_item loc;        // location in |E|
   list_item adj_loc;    // location in adjacencylist of source(e)
   list_item adj_loc2;   // location in adjacencylist of target(e) (ugraph)
   ent contents;         // user defined pointer type

   i_edge(node v, node w, ent i)
   { 
     contents=i;
     i_name=-1;
     s=v;
     t=w;
     loc=0;
     adj_loc=0;
     adj_loc2=0;
   }

   OPERATOR_NEW(7)  
   OPERATOR_DEL(7)

friend node   source(edge e) { return e->s;         }
friend node   target(edge e) { return e->t;         }
friend int    index(edge e)  { return e->i_name;    }

//space :  7 words = 28 bytes
};


//------------------------------------------------------------------------------
// friends of both nodes and edges
//------------------------------------------------------------------------------

inline graph* graph_of(edge e)           { return e->s->g;      }



typedef int  (*COMPARE(graph,node)) (node&, node&);
typedef int  (*COMPARE(graph,edge)) (edge&, edge&);

//------------------------------------------------------------------------------
// graph: base class for all graphs
//------------------------------------------------------------------------------

class graph {

friend class ugraph;
friend class planar_map;

list(node) V;              /* list of all nodes */
list(edge) E;              /* list of all edges */

int max_node_index;      // maximal node index 
int max_edge_index;      // maximal edge index

graph* parent;           // for subgraphs

bool   undirected;       // true iff graph undirected

/* space: 2 list + 4 word + "virtual" = 2*20 + 5*4 bytes = 60 bytes */

virtual void init_node_entry(ent& x)  const { x=0; }
virtual void init_edge_entry(ent& x)  const { x=0; }

virtual void copy_node_entry(ent& x)  const { x=x; }
virtual void copy_edge_entry(ent& x)  const { x=x; }

virtual void clear_node_entry(ent& x) const { x=0; }
virtual void clear_edge_entry(ent& x) const { x=0; }

virtual void read_node_entry(istream& in, ent& x)  const { Read(int(x),in); }
virtual void read_edge_entry(istream& in, ent& x)  const { Read(int(x),in); }

virtual void write_node_entry(ostream& o, ent& x)  const { Print(int(x),o); }
virtual void write_edge_entry(ostream& o, ent& x)  const { Print(int(x),o); }

virtual string node_type()  const { return "void"; }
virtual string edge_type()  const { return "void"; }

protected:

void copy_all_entries();

public:

   int  indeg(node v)     const   { return v->indegree; }
   int  outdeg(node v)    const   { return v->adj_edges.length(); }
   node source(edge e)    const   { return e->s; }
   node target(edge e)    const   { return e->t; }


   int  space() const;

   graph* super()         const   { return parent; }
   int max_node_i()       const   { return max_node_index; }
   int max_edge_i()       const   { return max_edge_index; }

   int  number_of_nodes() const   { return V.length(); }
   int  number_of_edges() const   { return E.length(); }

   list(node) all_nodes()       const { return V; }
   list(edge) all_edges()       const { return E; }
   list(edge) adj_edges(node v) const { return v->adj_edges; }
   list(node) adj_nodes(node)   const;

   ent& entry(node v)     { return v->contents; }
   ent& entry(edge e)     { return e->contents; }
   ent  inf(node v) const { return v->contents; }
   ent  inf(edge e) const { return e->contents; }

   void assign(node v,ent x) { v->contents = x; }
   void assign(edge e,ent x) { e->contents = x; }

   void reset() const;
   void init_node_iterator() const { V.init_iterator(); }
   void init_edge_iterator() const { E.init_iterator(); }
   int  next_node(node& v)   const { return V.next_element(v); }
   int  prev_node(node& v)   const { return V.prev_element(v); }
   int  next_edge(edge& e)   const { return E.next_element(e); }
   int  prev_edge(edge& e)   const { return E.prev_element(e); }

   edge first_adj_edge(node v)  const  { return v->adj_edges.head(); }
   edge last_adj_edge(node v)   const  { return v->adj_edges.tail(); }

   void init_adj_iterator(node v) const { v->adj_edges.init_iterator(); }

   int  next_adj_edge(edge& e,node v) const 
   { return v->adj_edges.next_element(e); }

   int  current_adj_edge(edge& e,node v) const 
   { return v->adj_edges.current_element(e);}


   int next_adj_node(node& w,node v)  const
   { edge e;
     if (next_adj_edge(e,v))
       { w = (v==e->s) ? e->t : e->s;
         return 1; }
     else { w = 0; return 0;}
   }
   
   int current_adj_node(node& w,node v)  const
   { edge e;
     if (current_adj_edge(e,v))
     { w = (v==e->s) ? e->t : e->s; return 1; }
     else { w = 0; return 0; }
   }
   
   node first_node()   const { return V.head(); }
   edge first_edge()   const { return E.head(); }
   node last_node()    const { return V.tail(); }
   edge last_edge()    const { return E.tail(); }

   node choose_node()  const { return V.head(); }
   edge choose_edge()  const { return E.head(); }

   node succ_node(node v)  const  { list_item loc = V.succ(v->loc);
                                    return (loc) ? V.inf(loc) : 0;
                                   }

   node pred_node(node v)  const  { list_item loc = V.pred(v->loc);
                                    return (loc) ? V.inf(loc) : 0;
                                   }

   edge succ_edge(edge e)  const  { list_item loc = E.succ(e->loc);
                                    return (loc) ? E.inf(loc) : 0;
                                   }

   edge pred_edge(edge e)  const  { list_item loc = E.pred(e->loc);
                                    return (loc) ? E.inf(loc) : 0;
                                   }


   edge adj_succ(edge e) const 
                         { list_item loc = e->s->adj_edges.succ(e->adj_loc);
                           return (loc) ? e->s->adj_edges[loc] : 0; } 

   edge adj_pred(edge e) const 
                         { list_item loc = e->s->adj_edges.pred(e->adj_loc);
                           return (loc) ? e->s->adj_edges[loc] : 0; } 

   edge cyclic_adj_succ(edge e) const 
   { list_item loc = e->s->adj_edges.cyclic_succ(e->adj_loc);
     return e->s->adj_edges[loc];
    }

   edge cyclic_adj_pred(edge e)  const
   { list_item loc = e->s->adj_edges.cyclic_pred(e->adj_loc);
     return e->s->adj_edges[loc];
    }

   node new_node(ent);

   node new_node()   { ent x; init_node_entry(x);  
                       return new_node(x); }

   void del_node(node);

   void del_edge(edge);

   edge new_edge(node, node, ent);

   edge new_edge(node v, node w) 
   { ent x; init_edge_entry(x);
     return new_edge(v,w,x);}

   edge new_edge(edge, node, ent, int dir);

   edge new_edge(edge e, node w, int dir=0) 
   { ent x; init_edge_entry(x);
     return new_edge(e,w,x,dir); }


   virtual void print_node(node v)  const
           { cout << form("[%d]",v->i_name); }

   virtual void print_edge(edge e)  const
           { cout << form("[%d]--->[%d]",e->s->i_name,e->t->i_name); }

   void del_all_nodes(); 
   void del_all_edges(); 

   list(edge) insert_reverse_edges();

   void sort_edges(COMPARE(graph,edge));
   void sort_nodes(COMPARE(graph,node) f ) { V.sort(f); }

   edge rev_edge(edge);
   graph& rev();

   void make_directed();
   void make_undirected();

   void write(string s) const;
   int  read (string s);

   void print(string s) const;
   void print()         const   { print("");}
   void clear();

   graph& operator=(const graph&); 

   graph() {max_node_index=max_edge_index=-1; parent = 0; undirected = false;}

   //subgraph constructors
   graph(graph&, list(node)&, list(edge)&);
   graph(graph&, list(edge)&);

   graph(const graph&);

   ~graph()     {  clear(); }


};


//------------------------------------------------------------------------------
// GRAPH: generic directed graphs
//------------------------------------------------------------------------------

#define GRAPH(vtype,etype)   name3(vtype,etype,GRAPH)

#define GRAPHdeclare2(vtype,etype)\
class GRAPH(vtype,etype) : public graph {\
\
void init_node_entry(ent& x) const  { x = Copy(vtype(0)); }\
void init_edge_entry(ent& x) const  { x = Copy(etype(0)); }\
\
void copy_node_entry(ent& x) const  { Copy(*(vtype*)&x); }\
void copy_edge_entry(ent& x) const  { Copy(*(etype*)&x); }\
\
void clear_node_entry(ent& x) const { Clear(*(vtype*)&x); }\
void clear_edge_entry(ent& x) const { Clear(*(etype*)&x); }\
\
void read_node_entry(istream& i, ent& x)  const { Read(*(vtype*)&x,i); }\
void read_edge_entry(istream& i, ent& x)  const { Read(*(etype*)&x,i); }\
\
void write_node_entry(ostream& o, ent& x)  const { Print(*(vtype*)&x,o); }\
void write_edge_entry(ostream& o, ent& x)  const { Print(*(etype*)&x,o); }\
\
string node_type()  const { return STRINGIZE(vtype); }\
string edge_type()  const { return STRINGIZE(etype); }\
\
public:\
\
   vtype  inf(node v)    const   { return vtype(graph::inf(v)); }\
   etype  inf(edge e)    const   { return etype(graph::inf(e)); }\
   vtype& operator[] (node v)         { return *(vtype*)&graph::entry(v); }\
   vtype  operator[] (node v) const   { return vtype(graph::inf(v)); }\
   etype& operator[] (edge e)         { return *(etype*)&graph::entry(e); }\
   etype  operator[] (edge e) const   { return etype(graph::inf(e)); }\
   void   assign(node v,vtype x) { graph::assign(v,Ent(x)); }\
   void   assign(edge e,etype x) { graph::assign(e,Ent(x)); }\
   node   new_node()             { return graph::new_node(); }\
   node   new_node(vtype a)      { return graph::new_node(Copy(a)); }\
   edge   new_edge(node v, node w)\
                              { return graph::new_edge(v,w); }\
   edge   new_edge(node v, node w, etype a)\
                              { return graph::new_edge(v,w,Copy(a)); }\
   edge   new_edge(edge e, node w)\
                              { return graph::new_edge(e,w); }\
   edge   new_edge(edge e, node w, etype a)\
                              { return graph::new_edge(e,w,Copy(a),0); }\
   edge   new_edge(edge e, node w, etype a, int dir)\
                              { return graph::new_edge(e,w,Copy(a),dir); }\
\
   void   print_node(node v) const\
                             { if(super() != 0)\
                                   super()->print_node(node(graph::inf(v)));\
                               else { cout << "[" << index(v) <<"] (" ;\
                                      Print(inf(v));\
                                      cout << ")";\
                                     }\
                              }\
\
   void   print_edge(edge e) const\
                            { if(super() != 0)\
                                   super()->print_edge(edge(graph::inf(e)));\
                              else { cout << form("[%d]--(",index(source(e)));\
                                     Print(inf(e));\
                                     cout << form(")-->[%d]",index(target(e)));\
                                    }\
                              }\
\
\
   GRAPH(vtype,etype)& operator=(const GRAPH(vtype,etype)& a)\
       {return (GRAPH(vtype,etype)&)(graph::operator=((graph&)a)); }\
   GRAPH(vtype,etype)()     {}\
   GRAPH(vtype,etype)(const GRAPH(vtype,etype)& a) : graph((graph&)a) { copy_all_entries();}\
/* sub graphs */\
   GRAPH(vtype,etype)(GRAPH(vtype,etype)& a, list(node)& b, list(edge)& c)\
                                                : graph((graph&)a,b,c) {}\
   GRAPH(vtype,etype)(GRAPH(vtype,etype)& a, list(edge)& c)\
                                                : graph((graph&)a,c) {}\
   ~GRAPH(vtype,etype)()     { clear(); }\
};


//------------------------------------------------------------------------------
// Macro zur Definition von Feldern mit Indextyp node oder edge
//------------------------------------------------------------------------------

// declaration (itype = node/edge):

#define graph_array(itype) name2(itype,graph_array)

#define graph_arraydeclare(itype)\
\
class graph_array(itype) {\
graph* g;  /* array is declared for graph *g */  \
int max_i; /* maximal node index in g at declaration time */  \
ent* arr;  \
virtual void clear_entry(ent& x) const { x=0; }\
virtual void copy_entry(ent& x)  const { x=x; }\
virtual void init_entry(ent& x)  const { x=0; }\
public:\
 ent& entry(itype);\
 ent  inf(itype) const;\
 void init(const graph&, ent);\
 void init(const graph_array(itype)&);\
 void clear();\
 ent& elem(int i)                    { return arr[i]; }\
 int  size()               const     { return max_i+1;}\
 graph_array(itype)()                { g=0; max_i=(-1); arr=0;}\
~graph_array(itype)()                { clear(); }\
};\


declare(graph_array,node)
declare(graph_array,edge)

//------------------------------------------------------------------------------
// node arrays
//------------------------------------------------------------------------------

#define node_array(TYPE) name2(TYPE,node_array)

#define node_arraydeclare(type)\
\
class node_array(type) : public graph_array(node) {\
void copy_entry(ent& x)  const { x = Copy(*(type*)&x);  }\
void clear_entry(ent& x) const { Clear(*(type*)&x);     }\
void init_entry(ent& x)  const { x = Copy(type(0)); }\
public:\
type& operator[](node x)       { return *(type*)&graph_array(node)::entry(x);}\
type  operator[](node x) const { return type(graph_array(node)::inf(x));}\
type& elem(int i)           { return *(type*)&graph_array(node)::elem(i);}\
void  init(const graph& G, type i){ graph_array(node)::init(G,Copy(i)); }\
void  init(const graph& G)  { type y=0; graph_array(node)::init(G,Copy(y)); }\
void  init(const node_array(type)& A)\
                            { graph_array(node)::init((graph_array(node)&) A); }\
\
node_array(type)()                          {}\
node_array(type)(const graph& G)            { init(G);   }\
node_array(type)(const graph& G, type i)    { init(G,i); }\
node_array(type)(const node_array(type)& A) { init(A);   }\
node_array(type)& operator=(const node_array(type)& A) { init(A); return *this;}\
\
~node_array(type)()                { clear(); }\
};


//------------------------------------------------------------------------------
// edge arrays
//------------------------------------------------------------------------------

#define edge_array(TYPE) name2(TYPE,edge_array)

#define edge_arraydeclare(type)\
\
class edge_array(type) : public graph_array(edge) {\
void copy_entry(ent& x)  const { x = Copy(*(type*)&x);  }\
void clear_entry(ent& x) const { Clear(*(type*)&x);     }\
void init_entry(ent& x)  const { x = Copy(type(0)); }\
public:\
type& operator[](edge x)       { return *(type*)&graph_array(edge)::entry(x);}\
type  operator[](edge x) const { return type(graph_array(edge)::inf(x));}\
type& elem(int i)           { return *(type*)&graph_array(edge)::elem(i);}\
void  init(const graph& G, type i){ graph_array(edge)::init(G,Copy(i)); }\
void  init(const graph& G)  { type y=0; graph_array(edge)::init(G,Copy(y)); }\
void  init(const edge_array(type)& A)\
                            { graph_array(edge)::init((graph_array(edge)&) A); }\
\
edge_array(type)()                          {}\
edge_array(type)(const graph& G)            { init(G);   }\
edge_array(type)(const graph& G, type i)    { init(G,i); }\
edge_array(type)(const edge_array(type)& A) { init(A);   }\
edge_array(type)& operator=(const edge_array(type)& A) { init(A); return *this;}\
\
~edge_array(type)()                { clear(); }\
};

//------------------------------------------------------------------------------
// node matrices
//------------------------------------------------------------------------------

typedef graph_array(node)* node_array_ptr;

declare(node_array,node_array_ptr);

class Node_Matrix {
graph* g;
node_array(node_array_ptr) M;
virtual void init_entry(ent& x)  const { x=0; }
virtual void copy_entry(ent& x)  const { x=x; }
virtual void clear_entry(ent& x) const { x=0; }
virtual void print_entry(ent& x) const { Print(x); }
public:
 graph_array(node)& operator[](node v);
 ent& entry(node, node);
 ent  inf(node, node) const;
 void clear();
 void init(const graph& G, ent x);
 void init(Node_Matrix& M);
 Node_Matrix()  {}
~Node_Matrix()  { clear(); }
};


#define node_matrix(type) name2(type,node_matrix)

#define node_matrixdeclare(type)\
\
class node_matrix(type): public Node_Matrix {\
void copy_entry(ent& x)  const { Copy(*(type*)&x);  }\
void clear_entry(ent& x) const { Clear(*(type*)&x); }\
void init_entry(ent& x)  const { x = Copy(type(0)); }\
void print_entry(ent& x) const { Print(*(type*)&x); }\
public:\
node_array(type)& operator[](node v)\
{ return *(node_array(type)*)&(Node_Matrix::operator[](v)); }\
\
type& operator()(node v, node w) { return *(type*)&(Node_Matrix::entry(v,w));}\
type  operator()(node v, node w) const { return type(Node_Matrix::inf(v,w)); }\
\
void  init(const graph& G, type i) { Node_Matrix::init(G,Copy(i)); }\
void  init(const graph& G)    { ent i; init_entry(i); Node_Matrix::init(G,i); }\
void  init(const node_matrix(type)& M) { Node_Matrix::init((Node_Matrix&) M); }\
\
node_matrix(type)() {}\
node_matrix(type)(const graph& G)              { init(G);   }\
node_matrix(type)(const graph& G, type x)      { init(G,x); }\
node_matrix(type)(const node_matrix(type)& M)  { init(M); }\
\
node_matrix(type)& operator=(const node_matrix(type)& M){init(M);return *this;}\
\
~node_matrix(type)()                 { clear();   }\
};\


//------------------------------------------------------------------------------
// Iterationsanweisungen (Macros)
//------------------------------------------------------------------------------

/*
#define forall_nodes(a,b) for ((b).init_node_iterator(); (b).next_node(a); )
#define forall_edges(a,b) for ((b).init_edge_iterator(); (b).next_edge(a); )
*/

#define forall_nodes(a,g) for (a=(g).first_node(); a; a=(g).succ_node(a) )

#define forall_edges(a,g) for (a=(g).first_edge(); a; a=(g).succ_edge(a) )

#define forall_adj_edges(a,b)\
for (graph_of(b)->init_adj_iterator(b); graph_of(b)->next_adj_edge(a,b); )

/*
#define forall_adj_edges(a,b)\
   for (a=graph_of(b)->first_adj_edge(b); a; a=graph_of(b)->adj_succ(a) ) */
 

#define forall_adj_nodes(a,b)\
for (graph_of(b)->init_adj_iterator(b); graph_of(b)->next_adj_node(a,b); )

#define forall_node_pairs(a,b,g)\
forall_nodes(a,g) for (b=(g).first_node(); b; b=(g).succ_node(b) )


// reverse

#define Forall_nodes(a,g) for (a=(g).last_node(); a; a=(g).pred_node(a) )

#define Forall_edges(a,g) for (a=(g).last_edge(); a; a=(g).pred_edge(a) )


//------------------------------------------------------------------------------
// node_set  
//------------------------------------------------------------------------------

class node_set {
graph* g;
list(node) L;
graph_array(node) A;
public:
void insert(node x)  { if (A.inf(x) == nil) A.entry(x) = Ent(L.append(x)); }
void del(node x)     { if (A.inf(x) != nil) 
                        { L.del(list_item(A.inf(x))); A.entry(x) = nil;} 
                      }
bool member(node x)      { return (A.inf(x) != nil); }
node choose()  const     { return L.head(); }
int  size()    const     { return L.length(); }
bool empty()   const     { return L.empty(); }
void clear()             { L.clear(); A.init(*g,nil); }
node_set(const graph& G) { g = (graph*)&G; A.init(G,nil);}
~node_set()              { L.clear(); A.clear(); }
};

//------------------------------------------------------------------------------
// edge_set  
//------------------------------------------------------------------------------

class edge_set {
graph* g;
list(edge) L;
graph_array(edge) A;
public:
void insert(edge x)  { if (A.inf(x) == nil) A.entry(x) = Ent(L.append(x)); }
void del(edge x)     { if (A.inf(x) != nil) 
                        { L.del(list_item(A.inf(x))); A.entry(x) = nil;} 
                      }
bool member(edge x)      { return (A.inf(x) != nil) ? true:false; }
edge choose()  const     { return L.head(); }
int  size()    const     { return L.length(); }
bool empty()   const     { return L.empty(); }
void clear()             { L.clear(); A.init(*g,nil); }
edge_set(const graph& G) { g = (graph*)&G; A.init(G,nil);}
~edge_set()              { L.clear(); A.clear(); }
};


//------------------------------------------------------------------------------
// node partitions 
//------------------------------------------------------------------------------

#include "Partition.h"

class node_partition {

graph_array(node) I;
partition P;

public:
node_partition(const graph& G);
~node_partition()                 { P.clear();  }   

int  same_block(node v, node w)   
{ return P.same_block(partition_item(I.inf(v)),partition_item(I.inf(w))); }

void union_blocks(node v, node w) 
{ P.union_blocks(partition_item(I.inf(v)),partition_item(I.inf(w))); }

node find(node v) { return node(P.contents(P.find(partition_item(I.inf(v))))); }
};


//------------------------------------------------------------------------------
// node priority queues
//------------------------------------------------------------------------------

#ifndef FHEAPH
#include "F_heap.h"
#endif

#define node_pq(itype) name2(itype,node_pq)\

#define node_pqdeclare(itype)\
\
class node_pq(itype) : private f_heap{\
\
int cmp(ent& x, ent& y)  const { return compare(*(itype*)&x,*(itype*)&y); }\
void print_key(ent& x)   const { Print(*(itype*)&x); }\
void print_inf(ent& x)   const { Print(*(node*)&x);  }\
void clear_key(ent& x)   const { Clear(*(itype*)&x); }\
void clear_inf(ent& x)   const { Clear(*(node*)&x);  }\
void copy_key(ent& x)    const { x = Copy(*(itype*)&x);  }\
void copy_inf(ent& x)    const { x = Copy(*(node*)&x);   }\
graph_array(node) I;\
\
public:\
 node_pq(itype)(const graph& G) { I.init(G,0); }\
~node_pq(itype)()               { I.clear(); }\
\
 void decrease_inf(node v, itype i)\
                      { f_heap::decrease_key(f_heap_item(I.inf(v)),Ent(i)); }\
 void insert(node v,itype i)\
                      { ent x=Ent(i);I.entry(v)=(ent)f_heap::insert(x,ent(v)); }\
 void del(node v)     { f_heap::del_item(f_heap_item(I.inf(v)));   }\
 node find_min()      { return (node)f_heap::inf(f_heap::find_min());   }\
 node del_min()       { node v = find_min(); f_heap::del_min(); return v; }\
 void clear()         { f_heap::clear(); }\
 int empty()          { return f_heap::empty(); }\
\
};

//-----------------------------------------------------------------------------
// ugraph: base class for all undirected graphs
//-----------------------------------------------------------------------------

class ugraph : public graph{

public:

edge new_edge(node,node,ent);

edge new_edge(node v, node w)
{ ent x; init_edge_entry(x);
  return new_edge(v,w,x);
 }

edge new_edge(node,node,edge,edge,ent,int=0,int=0);

edge new_edge(node v, node w, edge e1 ,edge e2, int dir1=0,int dir2=0)
{ ent x; init_edge_entry(x);
  return new_edge(v,w,e1,e2,x,dir1,dir2);
 }

/* done by graph::del_edge

void del_edge(edge e) { e->t->adj_edges.del(e->adj_loc2);
                        graph::(e);
                       }
*/

list(node) adj_nodes(node) const;

node opposite(node v, edge e)  const;

int  degree(node v)     const  { return v->adj_edges.length(); }

edge adj_succ(edge e,node v) const;
edge adj_pred(edge e,node v) const;
edge cyclic_adj_succ(edge e,node v) const;
edge cyclic_adj_pred(edge e,node v) const;

void print_edge(edge e) const { print_node(source(e));
                                cout << "====";
                                print_node(target(e)); 
                               }

void read(string s) { graph::read(s); make_undirected(); } 

ugraph()  { undirected = true; }

ugraph(const graph& a): graph(a) { undirected = true; make_undirected(); }

ugraph(const ugraph& a) : graph((graph&)a)  { undirected = true;  }

~ugraph() { /* ~graph does the job */ }

//subgraph constructors
ugraph(ugraph&, list(node)&, list(edge)&);
ugraph(ugraph&, list(edge)&);

ugraph& operator=(const ugraph& a) 
                               { return (ugraph&)graph::operator=((graph&)a); }

ugraph& operator=(const graph& a)  { graph::operator=(a); 
                                     make_undirected();
                                     return (ugraph&)(*this);
                                    }

};


//------------------------------------------------------------------------------
// UGRAPH: generic ugraphs
//------------------------------------------------------------------------------

#define UGRAPH(vtype,etype)  name3(vtype,etype,UGRAPH)
#define ungraph(vtype,etype) name3(vtype,etype,UGRAPH)

#define UGRAPHdeclare2(vtype,etype)\
\
class UGRAPH(vtype,etype) : public ugraph {\
\
void init_node_entry(ent& x) const { x = Copy(vtype(0)); }\
void init_edge_entry(ent& x) const { x = Copy(etype(0)); }\
\
void copy_node_entry(ent& x) const { Copy(*(vtype*)&x); }\
void copy_edge_entry(ent& x) const { Copy(*(etype*)&x); }\
\
void clear_node_entry(ent& x) const { Clear(*(vtype*)&x); }\
void clear_edge_entry(ent& x) const { Clear(*(etype*)&x); }\
\
public:\
   vtype  inf(node v)         const { return vtype(graph::inf(v)); }\
   etype  inf(edge e)         const { return etype(graph::inf(e)); }\
   vtype& operator[] (node v)       { return *(vtype*)&ugraph::entry(v); }\
   vtype  operator[] (node v) const { return vtype(ugraph::inf(v)); }\
   etype& operator[] (edge e)       { return *(etype*)&ugraph::entry(e); }\
   etype  operator[] (edge e) const { return etype(ugraph::inf(e)); }\
   void   assign(node v,vtype x) { graph::assign(v,Ent(x)); }\
   void   assign(edge e,etype x) { graph::assign(e,Ent(x)); }\
   node   new_node()          { return ugraph::new_node(); }\
   node   new_node(vtype a)   { return ugraph::new_node(Copy(a)); }\
   edge   new_edge(node v, node w)\
                              { return ugraph::new_edge(v,w); }\
   edge   new_edge(node v, node w, etype a)\
                              { return ugraph::new_edge(v,w,Copy(a)); }\
   edge   new_edge(node v, node w, edge e1, edge e2, int dir1, int dir2)\
                       { return ugraph::new_edge(v,w,e1,e2,dir1,dir2); }\
   edge   new_edge(node v,node w,edge e1,edge e2,etype a,int dir1,int dir2)\
                       { return ugraph::new_edge(v,w,e1,e2,Copy(a),dir1,dir2); }\
\
   void   print_node(node v) const { cout << "[" << index(v) <<"] (" ; \
                                     Print(inf(v));\
                                     cout << ")";\
                                    }\
\
   void   print_edge(edge e) const { cout << form("[%d]==(",index(source(e)));\
                                     Print(inf(e));\
                                     cout << form(")==[%d]",index(target(e)));\
                                    }\
\
   UGRAPH(vtype,etype)& operator=(const UGRAPH(vtype,etype)& a)\
       {return (UGRAPH(vtype,etype)&)(ugraph::operator=((ugraph&)a)); }\
\
   UGRAPH(vtype,etype)& operator=(const graph& a)\
       {return (UGRAPH(vtype,etype)&)(ugraph::operator=(a)); }\
\
   UGRAPH(vtype,etype)()                                       { }\
   UGRAPH(vtype,etype)(const UGRAPH(vtype,etype)& a): ugraph((ugraph&)a) { copy_all_entries();}\
   UGRAPH(vtype,etype)(const graph& a)              : ugraph(a)         { copy_all_entries();}\
\
   ~UGRAPH(vtype,etype)() { clear(); }\
};\



//-----------------------------------------------------------------------------
// Predefined:
//-----------------------------------------------------------------------------
                
declare(node_array,int)
declare(node_array,real)
declare(node_array,node)
declare(node_array,edge)

declare(edge_array,int)
declare(edge_array,real)
declare(edge_array,node)
declare(edge_array,edge)

declare(node_matrix,int)
declare(node_matrix,real)

typedef node_array(int) node_array(bool);
typedef edge_array(int) edge_array(bool);
typedef node_matrix(int) node_matrix(bool);



//-----------------------------------------------------------------------------
// miscellaneous
//-----------------------------------------------------------------------------

extern bool compute_correspondence(const graph&, edge_array(edge)&);
extern void eliminate_parallel_edges(graph& G);

extern void complete_graph(graph&,int);
extern void complete_graph(ugraph&,int);

extern void random_graph(graph&,int,int);
extern void random_graph(ugraph&,int,int);

extern void random_planar_graph(graph&,int);

extern void test_graph(graph&);
extern void test_graph(ugraph&);
extern void test_bigraph(graph&, list(node)&, list(node)&);

#endif


