/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * zoom.C - modulo de execucao de zoom. 
 *
 * 
 * ATENCAO:
 *         Lentes de aumento estao desabilitadas em estacoes coloridas
 *         pois sao muito lentas e ocupam memoria excessivamente.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    08.10.91  21:25 	Welson R. Jacometti  	Arquivo criado
 *    21.10.91  12:00	Welson			Lentes de aumento paradas
 *    22.10.91  11:05 	Welson			Lentes que se movimentam
 *
 */

#include "color.h"
#include "zoom.h"

/*
 * Variaveis exportaveis (via extern) que podem ser ajustadas via janela
 * de properties. 
 *
 */

double ZOOM_FACTOR	= 2;		// 100%
double GLASS1_FACTOR 	= 4;
double GLASS2_FACTOR	= 8;
bool ZOOM_CENTERED 	= false;
bool ZOOM_STACKED	= false;

/*
 * Constantes e variaveis internas ao modulo de zoom.
 *
 */
 
#define GLASS1_WIDTH 	125
#define GLASS1_HEIGHT	100
#define GLASS2_WIDTH	175
#define GLASS2_HEIGHT	150

short blank_cursor_icon[] = {
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
};

#define BLANK_CURSOR_WIDTH	16
#define BLANK_CURSOR_HEIGHT	16

bool glass_events(World *, Xv_Window, Event *, bool &);
bool zoom_in_out_events(World *, Xv_Window, Event *);

static GC gc;
static Server_image blank_cursor_image;
static Xv_Cursor blank_cursor;
static Canvas glass1, glass2;
static int ZOOM_INITIALIZED = false;
static int GLASSES_ALLOWED = false;

/*
 * zoom_init() - inicializa variaveis internas ao modulo de zoom
 *              
 *
 */

void zoom_init(World *world) {
    if (!ZOOM_INITIALIZED) {
      ZOOM_INITIALIZED = true;
      GLASSES_ALLOWED = (SCREEN_DEPTH == 1);
      
      if (GLASSES_ALLOWED) {
        XGCValues v;
      
        blank_cursor_image = (Server_image)xv_create(NULL, SERVER_IMAGE,
      				XV_WIDTH,		BLANK_CURSOR_WIDTH,
    				XV_HEIGHT,		BLANK_CURSOR_HEIGHT,
    				SERVER_IMAGE_BITS,	blank_cursor_icon,
    				NULL);
    
        blank_cursor = (Xv_Cursor)xv_create(NULL, CURSOR,
    				CURSOR_IMAGE,		blank_cursor_image,
        			NULL);

        v.foreground = (unsigned long)xv_get(world->get_canvas(), WIN_BACKGROUND_COLOR);
        gc = world->CreateGC(GCForeground, v);

        glass1 = (Canvas)xv_create(world->get_canvas(), CANVAS,
      	   			CANVAS_X_PAINT_WINDOW,	TRUE,
      				XV_WIDTH,		2 * GLASS1_WIDTH,
      				XV_HEIGHT,		2 * GLASS1_HEIGHT,
      				XV_SHOW,		FALSE,
      				NULL);
      					
        (void)xv_set(canvas_paint_window(glass1),
      				WIN_CURSOR,		blank_cursor,
      				WIN_CONSUME_EVENTS,	WIN_NO_EVENTS,
      							NULL,
      				NULL);
      		
        glass2 = (Canvas)xv_create(world->get_canvas(), CANVAS,
      	   			CANVAS_X_PAINT_WINDOW,	TRUE,
      				XV_WIDTH,		2 * GLASS2_WIDTH,
      				XV_HEIGHT,		2 * GLASS2_HEIGHT,
      				XV_SHOW,		FALSE,
      				NULL);
      					
        (void)xv_set(canvas_paint_window(glass2),
      				WIN_CURSOR,		blank_cursor,
      				WIN_CONSUME_EVENTS,	WIN_NO_EVENTS,
      							NULL,
      				NULL);      		
      }
    }
}

/*
 * zoom_reset() - executa `reset' no World especificado
 *
 */
 
void zoom_reset(World *world) {
    world->reset();
    world->ClearWindow();
    world->call_repaint_proc();
}

void zoom_in(World *world) {
    double x1, y1, x2, y2, x, y;
    
    world->get_values(x1, y1, x2, y2);
    x = (x1 + x2) / 2;
    y = (y1 + y2) / 2;
    
    if (ZOOM_STACKED)
      world->zoom_in_stack(x, y, ZOOM_FACTOR);
    else
      world->zoom_in(x, y, ZOOM_FACTOR);

    world->ClearWindow();
    world->call_repaint_proc();
}
        
void zoom_out(World *world) {
    double x1, y1, x2, y2, x, y;
    
    world->get_values(x1, y1, x2, y2);
    x = (x1 + x2) / 2;
    y = (y1 + y2) / 2;
    
    if (ZOOM_STACKED)
      world->zoom_out_stack(x, y, ZOOM_FACTOR);
    else
      world->zoom_out(x, y, ZOOM_FACTOR);

    world->ClearWindow();
    world->call_repaint_proc();
}

/*
 * zoom_events() - trata eventos ativando funcoes de zoom.
 *
 * Button LEFT   - zoom_in()  || zoom_in_stack()
 *        MIDDLE - zoom_out() || zoom_out_stack()
 *
 *   SHIFT LEFT   - glass #1
 *   SHIFT MIDDLE - glass #2
 *
 */

bool zoom_events(World *world, Xv_Window window, Event *event) {
    static bool show_glass = false;
    
    if (GLASSES_ALLOWED && (event_shift_is_down(event) && event_is_down(event) || show_glass)) 
      return glass_events(world, window, event, show_glass);
    else
      return zoom_in_out_events(world, window, event);
}

bool glass_events(World *world, Xv_Window, Event *event, bool &show_glass) {
    double x1, y1, x2, y2;
    int x, y, nx, ny;
    static World *wglass = NULL;
    static double glass_factor;
    static int glass_width, glass_height;
    static Canvas glass;

    if (!ZOOM_INITIALIZED) {
      cerr << "zoom.C: glass_events: zoom variables not initialized - init_zoom()" << endl;
      exit(1);
    }
      
    switch (event_action(event)) {
      case ACTION_SELECT:				
      case ACTION_ADJUST:
    	if (event_is_down(event) && !wglass) {
    	  if (event_action(event) == ACTION_SELECT) {
    	    glass_factor = GLASS1_FACTOR;
   	    glass_width  = GLASS1_WIDTH;
    	    glass_height = GLASS1_HEIGHT;
    	    glass        = glass1;
    	  }
    	  else {
    	    glass_factor = GLASS2_FACTOR;
    	    glass_width  = GLASS2_WIDTH;
    	    glass_height = GLASS2_HEIGHT;
    	    glass        = glass2;
    	  }

          world->get_values(x1, y1, x2, y2);
          world->zoom_in_stack((x1 + x2) / 2, (y1 + y2) / 2, glass_factor);
          
          (void)xv_set(canvas_paint_window(glass),
          			XV_WIDTH, 	world->xcoord(x2) - world->xcoord(x1),
          			XV_HEIGHT,	world->ycoord(y2) - world->ycoord(y1),
              			NULL);

          world->zoom_previous();
	  world->get_values(x1, y1, x2, y2);
          wglass = new World(glass, x1, y1, x2, y2);
          assert(wglass);

    	  x = event_x(event);
          y = event_y(event);

          nx = wglass->xcoord(world->virtual_xcoord(x)) - glass_width;
          ny = wglass->ycoord(world->virtual_ycoord(y)) - glass_height;
          
          wglass->set_repaint_proc(world->get_repaint_proc());
          wglass->call_repaint_proc();

   	  (void)xv_set(glass, XV_X,	x - glass_width,
   	  		      XV_Y,	y - glass_height,
   	  		      XV_SHOW,	TRUE,
   	  		      NULL);

          wglass->pscroll_begin(gc);
          wglass->pscroll_paint_bigger(gc, -nx, -ny);

          show_glass = true;
        }
        else 
          if (event_is_up(event) && wglass) {
            (void)xv_set(glass, XV_SHOW, FALSE, NULL);
      	    wglass->pscroll_end(0, 0);
      	    wglass->ClearWindow();
      	    delete wglass;
      	    wglass = NULL;
      	  }
      	show_glass = false;
        break;
         
      case LOC_DRAG:
        if (wglass) {
          x = event_x(event);
          y = event_y(event);

          nx = wglass->xcoord(world->virtual_xcoord(x)) - glass_width;
          ny = wglass->ycoord(world->virtual_ycoord(y)) - glass_height;
          
          wglass->pscroll_paint_bigger(gc, -nx, -ny);

   	  (void)xv_set(glass, XV_X,	x - glass_width,
   			      XV_Y,	y - glass_height,
   	  		      XV_SHOW,	TRUE,
   	  		      NULL);
   	  show_glass = true;
   	}
        break;
        
      default:
        show_glass = false;
        break;
    }
    
    return false;
}

bool zoom_in_out_events(World *world, Xv_Window window, Event *event) {
    double x1, y1, x2, y2;
    double x, y;

    if (ZOOM_CENTERED) {
      world->get_values(x1, y1, x2, y2);
      x = (x1 + x2) / 2;
      y = (y1 + y2) / 2;
    }
    else {
      x = world->virtual_xcoord(event_x(event));
      y = world->virtual_ycoord(event_y(event));
    }
 
    switch (event_action(event)) {
      case ACTION_SELECT:
        if (event_is_up(event)) {
          if (ZOOM_STACKED)
            world->zoom_in_stack(x, y, ZOOM_FACTOR);
          else 
            world->zoom_in(x, y, ZOOM_FACTOR);
          world->ClearWindow();
          world->call_repaint_proc();
          if (!ZOOM_CENTERED) {
            int width, height;
            world->get_view(width, height);
            (void)xv_set(window, WIN_MOUSE_XY, width / 2, height / 2, NULL);
          }
        }
        break;
        
      case ACTION_ADJUST:
        if (event_is_up(event)) {
          if (ZOOM_STACKED) 
            world->zoom_out_stack(x, y, ZOOM_FACTOR);
          else 
            world->zoom_out(x, y, ZOOM_FACTOR);
          world->ClearWindow();
          world->call_repaint_proc();
          if (!ZOOM_CENTERED) {
            int width, height;
            world->get_view(width, height);
            (void)xv_set(window, WIN_MOUSE_XY, width / 2, height / 2, NULL);
          }
        }
        break;
    }
    
    return false;
}

